/*
 *  linux/arch/arm/mach-pxa/colibri-displays.c
 *
 *  Generic Toradex Colibri Display functions
 *  Yury Georgievskiy <yury.georgievskiy@noser.com>
 *
 *  Copyright (C) 2011 Noser Engineering
 *
 *  2011-01-05: Yury Georgievskiy <yury.georgievskiy@noser.com>
 *              initial version
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <asm/mach-types.h>
#include <asm/mach/arch.h>
#include <asm/mach/irq.h>
#include <asm/sizes.h>

#include <linux/gpio.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/pwm_backlight.h>

#include <mach/colibri.h>
#include <mach/pxafb.h>

#include "devices.h"

int colibri_gpio_backlight = -1;

/*
  Following displays described here:
  1. CONFIG_COLIBRI_LCD_VGA
  2. CONFIG_COLIBRI_LCD_SVGA
  3. CONFIG_COLIBRI_LCD_LG_PHILIPS_LB064V02
  4. CONFIG_COLIBRI_LCD_SHARP_LQ043T3DX02
  5. CONFIG_COLIBRI_LCD_KYOCERA_THG057VGLAC
  6. CONFIG_COLIBRI_LCD_EDT_ET057011DH6
  7. CONFIG_COLIBRI_LCD_AUO_G104SN03V1
  8. CONFIG_COLIBRI_LCD_AMP_AM640480GTM
 */

#if defined(CONFIG_FB_PXA) || defined(CONFIG_FB_PXA_MODULE)
static void colibri_backlight_power(int on)
{
	gpio_set_value(colibri_gpio_backlight, on);
}

#ifdef CONFIG_COLIBRI_LCD_VGA
/* VGA connector (X24) on Colibri Evaluation Board, (X8) on Orchid */
static struct pxafb_mode_info colibri_vga_mode = {
	/* VGA */
	/* pixel clock period in picoseconds (25.18 MHz) */
	.pixclock	= 38000,
	.xres		= 640,
	.yres		= 480,
	.bpp		= 32,
	.depth		= 18,
	.hsync_len	= 64,	/* pixels horizontal sync (max) */
	.left_margin	= 48,	/* pixels left border */
	.right_margin	= 48,	/* pixels right border */
	.vsync_len	= 2,	/* lines vertical sync */
	.upper_margin	= 37,	/* lines top border */
	.lower_margin	= 17,	/* lines bottom border */
	.sync		= FB_SYNC_VERT_HIGH_ACT,
	.cmap_greyscale = 0,
};

static struct pxafb_mach_info colibri_vga_info = {
	.modes			= &colibri_vga_mode,
	.num_modes		= 1,
	.cmap_inverse		= 0,
	.cmap_static		= 0,
	.lcd_conn		= LCD_COLOR_TFT_18BPP,
	.pxafb_backlight_power	= colibri_backlight_power,
};
/* CONFIG_COLIBRI_LCD_VGA */
#elif defined(CONFIG_COLIBRI_LCD_SVGA)
/* VGA connector (X24) on Colibri Evaluation Board, (X8) on Orchid */
static struct pxafb_mode_info colibri_svga_mode = {
	/* SVGA */
	/* pixel clock period in picoseconds (25.18 MHz) */
	.pixclock	= 25000,
	.xres		= 800,
	.yres		= 600,
	.bpp		= 32,
	.depth		= 18,
	.hsync_len	= 64,	/* pixels horizontal sync (max) */
	.left_margin	= 48,	/* pixels left border */
	.right_margin	= 48,	/* pixels right border */
	.vsync_len	= 2,	/* lines vertical sync */
	.upper_margin	= 37,	/* lines top border */
	.lower_margin	= 17,	/* lines bottom border */
	.sync		= FB_SYNC_VERT_HIGH_ACT,
	.cmap_greyscale = 0,
};

static struct pxafb_mach_info colibri_svga_info = {
	.modes			= &colibri_svga_mode,
	.num_modes		= 1,
	.cmap_inverse		= 0,
	.cmap_static		= 0,
	.lcd_conn		= LCD_COLOR_TFT_18BPP,
	.pxafb_backlight_power	= colibri_backlight_power,
};
/* CONFIG_COLIBRI_LCD_SVGA */
#elif defined(CONFIG_COLIBRI_LCD_LG_PHILIPS_LB064V02)
/* LG.Philips LB064V02-A1 VGA LCD panel connected with 16 data lines to the LCD
   connector (X18) on the Toradex Colibri Evaluation Board. */
static struct pxafb_mode_info lg_philips_lb064v02_mode = {
	/* pixel clock period in picoseconds (25 MHz) */
	.pixclock	= 40000,
	.xres		= 640,
	.yres		= 480,
	.bpp		= 16,
	.hsync_len	= 2,	/* pixels horizontal sync */
	.left_margin	= 160,	/* pixels left border */
	.right_margin	= 2,	/* pixels right border */
	.vsync_len	= 45,	/* lines vertical sync */
	.upper_margin	= 0,	/* lines top border */
	.lower_margin	= 0,	/* lines bottom border */
	.sync		= (FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT),
	.cmap_greyscale = 0,
};

static struct pxafb_mach_info lg_philips_lb064v02_info = {
	.modes			= &lg_philips_lb064v02_mode,
	.num_modes		= 1,
	.cmap_inverse		= 0,
	.cmap_static		= 0,
	.lcd_conn		= LCD_COLOR_TFT_16BPP | LCD_PCLK_EDGE_FALL,
	.pxafb_backlight_power	= colibri_backlight_power,
};
/* CONFIG_COLIBRI_LCD_LG_PHILIPS_LB064V02 */
#elif defined(CONFIG_COLIBRI_LCD_SHARP_LQ043T3DX02)
/* This supports the Toradex 4.3" Sharp LQ043T3DX02 WQVGA touch TFT panel
   connected with 16 data lines to the generic display connector (X20 on the
   Toradex Colibri Evaluation Board, X1 on the Orchid) as follows:

	Display

	Toradex		Sharp		Description
	X20		X1
	1	o	1		GND
	2	|	2		L_PCLK
	3	|	3		L_LCLK
	4	|	4		L_FCLK
	5	o	5		GND
	6	o	8		LDD12/RED2
	7	|	9		LDD13/RED3
	8	|	10		LDD14/RED4
	9	o	11		LDD15/RED5
	10				LDD16
	11				LDD17
	12		12		GND
	13	o	14		LDD6/GREEN1
	14	|	15		LDD7/GREEN2
	15	|	16		LDD8/GREEN3
	16	|	17		LDD9/GREEN4
	17	o	18		LDD10/GREEN5
	18		7		LDD11/RED1
	19		19		GND
	20	o	21		LDD0/BLUE1
	21	|	22		LDD1/BLUE2
	22	|	23		LDD2/BLUE3
	23	|	24		LDD3/BLUE4
	24	o	25		LDD4/BLUE5
	25		13		LDD5/GREEN0
	26	o	26		GND
	27	|	27		L_BIAS
	28	|	28		+V_DISPLAY/VDD (+3V3 JP4 2-3)
	29	|	29		+V_DISPLAY/VDD (+3V3 JP4 2-3)
	30	|	30		R/L (GND JP6)
	31	|	31		U/D (GND JP5)
	32	|	32		GND
	33	|	33		DISP_SPARE1/TSMY
	34	|	34		DISP_SPARE2/TSMX
	35	|	35		DISP_SPARE3/TSPY
	36	|	36		DISP_SPARE4/TSPX
	37	|	37		BL_ON/DISPLAY_ON (+3V3)
	38	|	38		BL_GND
	39	|	39		BL_+5V
	40	o	40		BL_GND */
static struct pxafb_mode_info sharp_lq043t3dx02_mode = {
	/* pixel clock period in picoseconds */
	.pixclock	= 101936, /* was 111111 */
	.xres		= 480,
	.yres		= 272,
	.bpp		= 16,
//	.bpp		= 24, /* or 32 */
//	.depth		= 18,
	.hsync_len	= 41,	/* pixels horizontal sync */
	.left_margin	= 2,	/* pixels left border */
	.right_margin	= 2,	/* pixels right border */
	.vsync_len	= 10,	/* lines vertical sync */
	.upper_margin	= 2,	/* lines top border */
	.lower_margin	= 2,	/* lines bottom border */
	.sync		= 0,
	.cmap_greyscale = 0,
};

static struct pxafb_mach_info sharp_lq043t3dx02_info = {
	.modes			= &sharp_lq043t3dx02_mode,
	.num_modes		= 1,
	.cmap_inverse		= 0,
	.cmap_static		= 0,
	.lcd_conn		= LCD_COLOR_TFT_18BPP,
	.pxafb_backlight_power	= colibri_backlight_power,
};
/* CONFIG_COLIBRI_LCD_SHARP_LQ043T3DX02 */
#elif defined(CONFIG_COLIBRI_LCD_KYOCERA_THG057VGLAC)
/* Kyocera THG057VGLAC-H00 TFT panel connected with 16 data lines to the
   generic display connector X20 on the Toradex Colibri Evaluation Board */
static struct pxafb_mode_info kyocera_thg057vglac_mode = {
	/* pixel clock period in picoseconds (25.18 MHz) */
	.pixclock	= 39714,
	.xres		= 640,
	.yres		= 480,
	.bpp		= 16,
	.hsync_len	= 64, /* pixels horizontal sync (max) */
	.left_margin	= 105, /* pixels left border */
	.right_margin	= 1, /* pixels right border */
	.vsync_len	= 17, /* lines vertical sync */
	.upper_margin	= 18, /* lines top border */
	.lower_margin	= 0, /* lines bottom border */
	.sync		= 0,
	.cmap_greyscale = 0,
};

static struct pxafb_mach_info kyocera_thg057vglac_info = {
	.modes			= &kyocera_thg057vglac_mode,
	.num_modes		= 1,
	.cmap_inverse		= 0,
	.cmap_static		= 0,
	.lcd_conn		= LCD_COLOR_TFT_16BPP,
	.pxafb_backlight_power	= colibri_backlight_power,
};
/* CONFIG_COLIBRI_LCD_KYOCERA_THG057VGLAC */
#elif defined(CONFIG_COLIBRI_LCD_EDT_ET057011DH6)
/* This supports the EDT ET057011DH6 QVGA TFT panel connected with 16 data
   lines. */
static struct pxafb_mode_info edt_et057011dh6_mode = {
	/* pixel clock period in picoseconds */
	.pixclock	= 40000,
	.xres		= 320,
	.yres		= 240,
	.bpp		= 16,
	.hsync_len	= 30, /* pixels horizontal sync (max) */
	.left_margin	= 20, /* pixels left border */
	.right_margin	= 38, /* pixels right border */
	.vsync_len	= 3, /* lines vertical sync */
	.upper_margin	= 5, /* lines top border */
	.lower_margin	= 15, /* lines bottom border */
	.sync		= (FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT),
	.cmap_greyscale = 0,
};

static struct pxafb_mach_info edt_et057011dh6_info = {
	.modes			= &edt_et057011dh6_mode,
	.num_modes		= 1,
	.cmap_inverse		= 0,
	.cmap_static		= 0,
	.lcd_conn		= LCD_COLOR_TFT_16BPP,
	.pxafb_backlight_power	= colibri_backlight_power,
};
 /* CONFIG_COLIBRI_LCD_EDT_ET057011DH6 */
#elif defined(CONFIG_COLIBRI_LCD_AUO_G104SN03V1)
/* This supports the AUO G104SN03 V1 10.4" SVGA LVDS TFT panel connected via
   Toradex LVDS Adapter with 1mm pitch to the generic display connector (X20 on
   the Toradex Colibri Evaluation Board, X1 on the Orchid). The 16 data lines
   are wired through the function connectors from X9 (X8 Row A) to X7
   (X8 Row B) resp. X12 (X11 Row A) as follows:

								X7 (X8 Row B)
	X9 (X8 Row A)	PXA Function	Display Function	X12 (X11 Row A)

	A10		LDD7/GREEN2	GREEN2			B19
	A11		LDD9/GREEN4	GREEN4			B26
	A12		LDD11/RED0	RED1			B14
	A13		LDD12/RED1	RED2			B21
	A14		LDD13/RED2	RED3			B20
	A17		LDD3/BLUE3	BLUE4			B28
	A18		LDD2/BLUE2	BLUE3			B17
	A19		LDD8/GREEN3	GREEN3			B11
	A20		LDD15/RED4	RED5			A24
	A21		LDD14/RED3	RED4			A21
	A24		LDD1/BLUE1	BLUE2			B18
	A25		LDD5/GREEN0	GREEN0			B30
	A26		LDD10/GREEN5	GREEN5			B12
	A27		LDD0/BLUE0	BLUE1			B24
	A28		LDD4/BLUE4	BLUE5			B25
	A30		LDD6/GREEN1	GREEN1			B10

   Remember to remove the respecitve jumpers on X8 resp. X11 as well. */
static struct pxafb_mode_info auo_g104sn03v1_mode = {
	/* SVGA */
	/* pixel clock period in picoseconds (40 MHz) */
	.pixclock	= 25000,
	.xres		= 800,
	.yres		= 600,
	.bpp		= 16,
//	.bpp		= 24, /* or 32 */
//	.depth		= 18,
	.hsync_len	= 64,	/* pixels horizontal sync (max) */
	.left_margin	= 117,	/* pixels left border */
	.right_margin	= 129,	/* pixels right border */
	.vsync_len	= 2,	/* lines vertical sync */
	.upper_margin	= 56,	/* lines top border */
	.lower_margin	= 36,	/* lines bottom border */
	.sync		= (FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT),
};

static struct pxafb_mach_info auo_g104sn03v1_info = {
	.modes			= &auo_g104sn03v1_mode,
	.num_modes		= 1,
	.lcd_conn		= LCD_COLOR_TFT_16BPP | LCD_PCLK_EDGE_FALL,
	.pxafb_backlight_power	= colibri_backlight_power,
};
/* CONFIG_COLIBRI_LCD_AUO_G104SN03V1 */
#elif defined (CONFIG_COLIBRI_LCD_AMP_AM640480GTM)
static struct pxafb_mode_info amp_am640480gtm_mode = {
	/* pixel clock period in picoseconds (40 MHz) */
	.pixclock	= 25000,
	.xres		= 640,
	.yres		= 480,
	.bpp		= 32,
	.depth		= 18,
	.hsync_len	= 64,	/* pixels horizontal sync (max) */
	.left_margin	= 48,	/* pixels left border */
	.right_margin	= 48,	/* pixels right border */
	.vsync_len	= 2,	/* lines vertical sync */
	.upper_margin	= 37,	/* lines top border */
	.lower_margin	= 17,	/* lines bottom border */
	.sync		= FB_SYNC_VERT_HIGH_ACT | FB_SYNC_HOR_HIGH_ACT,
	.cmap_greyscale = 0,
};

static struct pxafb_mach_info amp_am640480gtm_info = {
	.modes			= &amp_am640480gtm_mode,
	.num_modes		= 1,
	.cmap_inverse		= 0,
	.cmap_static		= 0,
	.lcd_conn		= LCD_COLOR_TFT_18BPP,
	.pxafb_backlight_power	= colibri_backlight_power,
};

static struct platform_pwm_backlight_data amp_am640480gtm_backlight_data = {
	.pwm_id         = 0,
	.max_brightness = 1023,
	.dft_brightness = 512,
	.pwm_period_ns  = 1000000,
};

#endif	/* CONFIG_COLIBRI_LCD_AMP_AM640480GTM */

static struct platform_device backlight_device = {
	.name   = "pwm-backlight",
	.dev    = {
		.parent         = &pxa27x_device_pwm0.dev,
	},
};

/**
 * @brief Request backlight gpio
 *
 * @param bl_pin -- gpio pin.
 *
 * @return void
 */
void __init colibri_pxa_init_lcd(int bl_pin)
{
	/* backlight GPIO: output, default on */
	colibri_gpio_backlight = bl_pin;
	gpio_request(bl_pin, "LCD backlight");
	gpio_direction_output(bl_pin, 1);

#ifdef CONFIG_COLIBRI_LCD_VGA
	set_pxa_fb_info(&colibri_vga_info);
#elif defined(CONFIG_COLIBRI_LCD_SVGA)
	set_pxa_fb_info(&colibri_svga_info);
#elif defined(CONFIG_COLIBRI_LCD_LG_PHILIPS_LB064V02)
	set_pxa_fb_info(&lg_philips_lb064v02_info);
#elif defined(CONFIG_COLIBRI_LCD_SHARP_LQ043T3DX02)
	set_pxa_fb_info(&sharp_lq043t3dx02_info);
#elif defined(CONFIG_COLIBRI_LCD_KYOCERA_THG057VGLAC)
	set_pxa_fb_info(&kyocera_thg057vglac_info);
#elif defined(CONFIG_COLIBRI_LCD_EDT_ET057011DH6)
	set_pxa_fb_info(&edt_et057011dh6_info);
#elif defined(CONFIG_COLIBRI_LCD_AUO_G104SN03V1)
	set_pxa_fb_info(&auo_g104sn03v1_info);
#elif defined(CONFIG_COLIBRI_LCD_AMP_AM640480GTM)
	set_pxa_fb_info(&amp_am640480gtm_info);
	backlight_device.dev.platform_data = &amp_am640480gtm_backlight_data;
#endif
	platform_device_register(&backlight_device);
}
#endif /* CONFIG_FB_PXA || CONFIG_FB_PXA_MODULE */
