/*
 *  linux/arch/arm/mach-pxa/colibri-pxa270.c
 *
 *  Support for Toradex PXA270 based Colibri module
 *  Daniel Mack <daniel@caiaq.de>
 *
 * Extensions copyright (C) 2008-2010 Noser Engineering
 *
 * 2008-07-21: Marcel Ziswiler <marcel.ziswiler@noser.com>
 *             initial version
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/sysdev.h>
#include <linux/interrupt.h>
#include <linux/bitops.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/physmap.h>
#include <linux/gpio.h>
#include <asm/mach-types.h>
#include <mach/hardware.h>
#include <asm/irq.h>
#include <asm/sizes.h>
#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>
#include <asm/mach/flash.h>

#include <plat/i2c.h>
#include <mach/pxa27x.h>
#include <mach/colibri.h>
#include <mach/mmc.h>

#include <mach/pxa2xx-regs.h>
#include <mach/pxafb.h>
#include <mach/audio.h>

#include <linux/ucb1400.h>
#include <linux/can/platform/sja1000.h>
#include <mach/mfp-pxa27x.h>

/* TODO: fixup_colibri() really required? */
#include <asm/setup.h>

#include "generic.h"
#include "devices.h"
#include "colibri.h"

#define COLIBRI_PXA270_ETH_PHYS		(PXA_CS2_PHYS)  /* Ethernet */
#define COLIBRI_PXA270_ETH_IRQ_GPIO	114
#define COLIBRI_PXA270_ETH_IRQ		\
	gpio_to_irq(mfp_to_gpio(COLIBRI_PXA270_ETH_IRQ_GPIO))

/*
 * GPIO configuration
 */
static mfp_cfg_t colibri_pxa270_pin_config[] __initdata = {
#if defined(CONFIG_SND_PXA2XX_AC97) || defined(CONFIG_SND_PXA2XX_AC97_MODULE)
	/* AC97 */
	GPIO28_AC97_BITCLK,
	GPIO29_AC97_SDATA_IN_0,
	GPIO30_AC97_SDATA_OUT,
	GPIO31_AC97_SYNC,
	GPIO95_AC97_nRESET,
	GPIO98_AC97_SYSCLK,
#endif /* CONFIG_SND_PXA2XX_AC97 || CONFIG_SND_PXA2XX_AC97_MODULE */

	GPIO113_GPIO,	/* Touchscreen interrupt */

#if defined(CONFIG_PCMCIA_PXA2XX) || defined(CONFIG_PCMCIA_PXA2XX_MODULE)
	/* CompactFlash (PCMCIA) */
	GPIO1_GPIO,		/* PRDY (GPIO_PCMCIA0_RDY_IRQ) */
	GPIO18_RDY,
	GPIO48_nPOE,
	GPIO49_nPWE,
	GPIO50_nPIOR,
	GPIO51_nPIOW,
	GPIO53_GPIO,		/* PRESET (GPIO_PCMCIA0_RESET) */
	GPIO54_nPCE_2,
	GPIO55_nPREG,
	GPIO56_nPWAIT,
	GPIO57_nIOIS16,
	GPIO82_GPIO,		/* PBVD2 (GPIO_PCMCIA0_BVD2) */
	GPIO83_GPIO,		/* PBVD1 (GPIO_PCMCIA0_BVD1) */
	GPIO84_GPIO,		/* PCD (GPIO_PCMCIA0_CD_IRQ) */
	GPIO85_nPCE_1,
	GPIO104_PSKTSEL,
	GPIO107_GPIO,		/* PPEN (GPIO_PCMCIA_POW_EN) */
#endif

	/* Ethernet */
	GPIO78_nCS_2,	/* Ethernet CS */
	GPIO114_GPIO,	/* Ethernet IRQ */

	/* 2nd Ethernet on Protea */
	GPIO79_nCS_3,
	GPIO83_GPIO,		/* Interrupt */

	/* External chip selects (e.g. CAN) */
	GPIO15_nCS_1,
	GPIO52_GPIO,		/* Interrupt */

	/* I2C */
#if defined(CONFIG_I2C_PXA) || defined(CONFIG_I2C_PXA_MODULE)
	GPIO117_I2C_SCL,
	GPIO118_I2C_SDA,
#endif /* CONFIG_I2C_PXA || CONFIG_I2C_PXA_MODULE */

	/* LEDs */
#if defined(CONFIG_LEDS_PWM) || defined(CONFIG_LEDS_PWM_MODULE)
	GPIO12_PWM3_OUT, /* SODIMM pin 59, EvalBoard X11 A23 */
	GPIO11_PWM2_OUT, /* SODIMM pin 28, EvalBoard X8 A2 */

#endif /* CONFIG_LEDS_PWM || CONFIG_LEDS_PWM_MODULE */
#if defined(CONFIG_LEDS_GPIO) || defined(CONFIG_LEDS_GPIO_MODULE)
	/* borrowed from PS2Top-Clock and PS2Bottom-Clock
	   as they are not widely used nowdays */
	GPIO14_GPIO,		/* SODIMM pin 63, EvalBoard X11 A25 */
	GPIO20_GPIO,		/* SODIMM pin 69, EvalBoard X11 A28 */
#endif /* CONFIG_LEDS_GPIO || CONFIG_LEDS_GPIO_MODULE */

#if defined(CONFIG_MMC_PXA) || defined(CONFIG_MMC_PXA_MODULE)
	/* MMC */
	GPIO0_GPIO,		/* Card detect interrupt */
	GPIO32_MMC_CLK,
	GPIO92_MMC_DAT_0,
	GPIO109_MMC_DAT_1,
	GPIO110_MMC_DAT_2,
	GPIO111_MMC_DAT_3,
	GPIO112_MMC_CMD,
#endif /* CONFIG_MMC_PXA || CONFIG_MMC_PXA_MODULE */

	/* UARTs */

	/* FFUART */
	GPIO10_FFUART_DCD,
	GPIO27_FFUART_RTS,
	GPIO33_FFUART_DSR,
	GPIO34_FFUART_RXD,
	GPIO38_FFUART_RI,
	GPIO39_FFUART_TXD,
	GPIO40_FFUART_DTR,
	GPIO100_FFUART_CTS,

	/* BTUART */
	GPIO42_BTUART_RXD,
	GPIO43_BTUART_TXD,
	GPIO44_BTUART_CTS,
	GPIO45_BTUART_RTS,

	/* STUART */
	GPIO46_STUART_RXD,
	GPIO47_STUART_TXD,

	/* USB */
//OHCI is only the USB host part of it
#if defined(CONFIG_USB_OHCI_HCD) || defined(CONFIG_USB_OHCI_HCD_MODULE)
	GPIO41_GPIO,		/* USB device cable detect, active high */
	GPIO88_USBH1_PWR,
	GPIO89_USBH1_PEN,
	/* Contrary to what the Marvell PXA27x Processor Developers Manual
	   Rev. A (November 3, 2006) states in note 6 on page 24-9 GPIO119
	   and GPIO120 seem to not only be available on Marvel PXA271,
	   PXA272 and PXA273. */
	GPIO119_USBH2_PWR,
	GPIO120_USBH2_PEN,
#endif /* CONFIG_USB_OHCI_HCD || CONFIG_USB_OHCI_HCD_MODULE */

#if defined(CONFIG_KEYBOARD_GPIO) || defined(CONFIG_KEYBOARD_GPIO_MODULE)
	/* Keys for Android
	   Note: the Toradex Colibri Evaluation Board will require jumper cables from
	   X11 (GPIO Usage) to X21 (LED/Switches) wired as follows
	   Key		GPIO	X11	X21
	   Home		36	B44	15	SWITCH8
	   Menu		35	B48	19	BUTTON2
	   Back		37	B47	17	BUTTON1 */
	GPIO36_GPIO | MFP_LPM_EDGE_RISE, /* Home */
	GPIO35_GPIO | MFP_LPM_EDGE_RISE, /* Menu */
	GPIO37_GPIO | MFP_LPM_EDGE_RISE, /* Back */
#endif /* CONFIG_KEYBOARD_GPIO || CONFIG_KEYBOARD_GPIO_MODULE */
};

/*
 * NOR flash defined in drivers/mtd/maps/colibri.c
 */

#define CONFIG_CAN_SJA1000 0 /*TODO create Kconfig entry */
#if defined(CONFIG_CAN_SJA1000) || defined(CONFIG_CAN_SJA1000_MODULE)
#define CAN_INTERRUPT 52 /*TODO solve for all colibri modules */
static struct resource colibri_can_resource[] = {
                        [0] =   {
                                .start  = PXA_CS1_PHYS,
                                .end    = PXA_CS1_PHYS+0xFF,
                                .flags  = IORESOURCE_MEM,
                                },
                        [1] =   {
                                .start  = gpio_to_irq(CAN_INTERRUPT),
                                .end    = gpio_to_irq(CAN_INTERRUPT),
                                .flags  = IORESOURCE_IRQ | IORESOURCE_IRQ_LOWEDGE,
                                }
};

static struct sja1000_platform_data colibri_can_platdata = {
        .osc_freq = (24000000 / 2),
        .ocr      = (OCR_MODE_NORMAL | OCR_TX0_PUSHPULL),
        .cdr      = CDR_CBP,
};

static struct platform_device colibri_can_device = {
        .name="sja1000_platform",
        .id=0,
        .num_resources=ARRAY_SIZE(colibri_can_resource),
        .resource=colibri_can_resource,
        .dev            = {
                .platform_data = &colibri_can_platdata
        }

};
#endif

#if defined(CONFIG_AX88796) || defined(CONFIG_AX88796_MODULE)
/* ASIX AX88796B 10/100 Ethernet controller */
static struct resource protea_asix_resource[] = {
	[0] = {
		.start	= PXA_CS3_PHYS,
		/* 2 means address lines shifted by one */
		.end	= PXA_CS3_PHYS + (0x20 * 2) - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		/* FIFO_SEL connected to address line 11 */
		.start	= PXA_CS3_PHYS + (1 << 11),
		.end	= PXA_CS3_PHYS + (1 << 11),
		.flags	= IORESOURCE_MEM,
	},
	[2] = {
		.start	= gpio_to_irq(mfp_to_gpio(MFP_PIN_GPIO83)),
		.end	= gpio_to_irq(mfp_to_gpio(MFP_PIN_GPIO83)),
		.flags	= IORESOURCE_IRQ | IORESOURCE_IRQ_LOWEDGE,
	}
};

static struct ax_plat_data protea_asix_platdata = {
	.flags		= AXFLG_MAC_FROMDEV,
	.wordlength	= 2,
};

static struct platform_device protea_device_asix = {
	.name		= "ax88796",
	.id		= -1,
	.num_resources	= ARRAY_SIZE(protea_asix_resource),
	.resource	= protea_asix_resource,
	.dev		= {
		.platform_data = &protea_asix_platdata
	}
};
#endif /* CONFIG_AX88796 || CONFIG_AX88796_MODULE */

/*
 * DM9000 Ethernet
 */
#if defined(CONFIG_DM9000) || defined(CONFIG_DM9000_MODULE)
static struct resource dm9000_resources[] = {
	[0] = { /* address port */
		.start	= COLIBRI_PXA270_ETH_PHYS,
		.end	= COLIBRI_PXA270_ETH_PHYS + 3,
		.flags	= IORESOURCE_MEM,
	},
	[1] = { /* data port (addr2) */
		.start	= COLIBRI_PXA270_ETH_PHYS + 4,
		.end	= COLIBRI_PXA270_ETH_PHYS + 4 + 500,
		.flags	= IORESOURCE_MEM,
	},
	[2] = {
		.start	= COLIBRI_PXA270_ETH_IRQ,
		.end	= COLIBRI_PXA270_ETH_IRQ,
		.flags	= IORESOURCE_IRQ | IORESOURCE_IRQ_HIGHEDGE
	},
};

static struct platform_device dm9000_device = {
	.name		= "dm9000",
	.id		= -1,
	.num_resources	= ARRAY_SIZE(dm9000_resources),
	.resource	= dm9000_resources,
};
#endif /* CONFIG_DM9000 || CONFIG_DM9000_MODULE */

#if defined(CONFIG_FB_PXA) || defined(CONFIG_FB_PXA_MODULE)
static mfp_cfg_t colibri_pxa270_lcd_pin_config[] __initdata = {
	GPIO16_PWM0_OUT,  /* PWM backlight */
	GPIO58_LCD_LDD_0,
	GPIO59_LCD_LDD_1,
	GPIO60_LCD_LDD_2,
	GPIO61_LCD_LDD_3,
	GPIO62_LCD_LDD_4,
	GPIO63_LCD_LDD_5,
	GPIO64_LCD_LDD_6,
	GPIO65_LCD_LDD_7,
	GPIO66_LCD_LDD_8,
	GPIO67_LCD_LDD_9,
	GPIO68_LCD_LDD_10,
	GPIO69_LCD_LDD_11,
	GPIO70_LCD_LDD_12,
	GPIO71_LCD_LDD_13,
	GPIO72_LCD_LDD_14,
	GPIO73_LCD_LDD_15,
	GPIO74_LCD_FCLK,
	GPIO75_LCD_LCLK,
	GPIO76_LCD_PCLK,
	GPIO77_LCD_BIAS,
	GPIO81_GPIO,		/* Backlight */
	GPIO86_LCD_LDD_16,
	GPIO87_LCD_LDD_17,
};

void __init colibri_pxa270_init_lcd(void)
{
	pxa2xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa270_lcd_pin_config));
}

#else
static inline void colibri_pxa270_init_lcd(void) {}
#endif /* CONFIG_FB_PXA || CONFIG_FB_PXA_MODULE */

/******************************************************************************
 * Audio and Touchscreen
 ******************************************************************************/
#if	defined(CONFIG_TOUCHSCREEN_UCB1400) || \
	defined(CONFIG_TOUCHSCREEN_UCB1400_MODULE)
static pxa2xx_audio_ops_t colibri_pxa270_ac97_pdata = {
	.reset_gpio	= -1
};

static struct ucb1400_pdata colibri_pxa270_ucb1400_pdata = {
	.irq = gpio_to_irq(mfp_to_gpio(MFP_PIN_GPIO113)),
};

static struct platform_device colibri_pxa270_ucb1400_device = {
	.name		= "ucb1400_core",
	.id		= -1,
	.dev		= {
		.platform_data = &colibri_pxa270_ucb1400_pdata,
	},
};

static void __init colibri_pxa270_tsc_init(void)
{
	/* Old NXP (formerly Philips) UCB1400 is used */
	pxa_device_ac97.name = "pxa2xx-ac97-ucb1400";

	pxa_set_ac97_info(&colibri_pxa270_ac97_pdata);
	platform_device_register(&colibri_pxa270_ucb1400_device);
}
#else
static inline void colibri_pxa270_tsc_init(void) {}
#endif

#if defined(CONFIG_MMC_PXA) || defined(CONFIG_MMC_PXA_MODULE)
static struct pxamci_platform_data colibri_mci_platform_data = {
	.detect_delay_ms  = 20,
	.ocr_mask	  = MMC_VDD_32_33|MMC_VDD_33_34,
	.gpio_card_ro     = -1,
	.gpio_card_detect = mfp_to_gpio(MFP_PIN_GPIO0),
	.gpio_power       = -1
};
#endif /* CONFIG_MMC_PXA || CONFIG_MMC_PXA_MODULE */

static struct platform_device *colibri_pxa270_devices[] __initdata = {
#if defined(CONFIG_DM9000)
	&dm9000_device,
#endif
};

static void __init colibri_pxa270_init(void)
{
	pxa2xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa270_pin_config));
	colibri_pxa270_tsc_init();
	pxa_set_ffuart_info(NULL); /* full feature */
	pxa_set_btuart_info(NULL); /* blue tooth */
	pxa_set_stuart_info(NULL); /* standart */
	colibri_pxa_init_evalboard();
	colibri_pxa_init_gpio_keys(PXA_ALL_CB);

	colibri_pxa270_init_lcd();
	colibri_pxa_init_lcd(mfp_to_gpio(MFP_PIN_GPIO81));
	colibri_pxa_init_rtc();

		/*
		 * Note: We depend that the bootloader set
		 * the correct value to MSC register for Ethernet.
		 */
#if defined(CONFIG_AX88796) || defined(CONFIG_AX88796_MODULE)
		platform_device_register(&protea_device_asix);
#endif /* CONFIG_AX88796 || CONFIG_AX88796_MODULE */

#if defined(CONFIG_MMC_PXA) || defined(CONFIG_MMC_PXA_MODULE)
	/* MMC card detect for controller 0 */
	colibri_pxa_init_mmc(&colibri_mci_platform_data);
#endif	/* CONFIG_MMC_PXA || CONFIG_MMC_PXA_MODULE */

	colibri_pxa_init_ohci(mfp_to_gpio(MFP_PIN_GPIO41));

#if defined(CONFIG_CAN_SJA1000) || defined(CONFIG_CAN_SJA1000_MODULE)
        platform_device_register(&colibri_can_device);
#endif

	platform_add_devices(ARRAY_AND_SIZE(colibri_pxa270_devices));
}

/* TODO: really required? */
static void __init
fixup_colibri(struct machine_desc *desc, struct tag *tags,
	      char **cmdline, struct meminfo *mi)
{
	SET_BANK(0, 0xa0000000, 64*1024*1024);
	mi->nr_banks = 1;
}

static void __init colibri_map_io(void)
{
	pxa_map_io();

	/* setup sleep mode values */
	PWER  = 0x00000002;
	PFER  = 0x00000000;
	PRER  = 0x00000002;
	PGSR0 = 0x00008000;
	PGSR1 = 0x003F0202;
	PGSR2 = 0x0001C000;
	PCFR |= PCFR_OPDE;
}

MACHINE_START(COLIBRI, "Toradex Colibri PXA270")
	.phys_io	= 0x40000000,
	.io_pg_offst	= (io_p2v(0x40000000) >> 18) & 0xfffc,
	.boot_params	= COLIBRI_SDRAM_BASE + 0x100,
	.fixup          = fixup_colibri,
	.init_machine	= colibri_pxa270_init,
	.map_io		= colibri_map_io,
	.init_irq	= pxa27x_init_irq,
	.timer		= &pxa_timer,
MACHINE_END

