/*
 *  arch/arm/mach-pxa/colibri-pxa300.c
 *
 *  Support for Toradex PXA300/310 based Colibri module
 *
 *  Daniel Mack <daniel@caiaq.de>
 *  Matthias Meier <matthias.j.meier@gmx.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <linux/interrupt.h>

#include <asm/mach-types.h>
#include <asm/sizes.h>
#include <asm/mach/arch.h>
#include <asm/mach/irq.h>

#include <mach/pxa300.h>
#include <mach/colibri.h>
#include <mach/pxafb.h>
#include <mach/audio.h>
#include <mach/mfp-pxa300.h>
#include <mach/mmc.h>

#include "generic.h"
#include "devices.h"
#include "colibri.h"

static mfp_cfg_t colibri_pxa300_pin_config[] __initdata = {
	/* I2C */
#if defined(CONFIG_I2C_PXA) || defined(CONFIG_I2C_PXA_MODULE)
	GPIO21_I2C_SCL | MFP_LPM_EDGE_BOTH,
	GPIO22_I2C_SDA | MFP_LPM_EDGE_BOTH,
#endif /* CONFIG_I2C_PXA || CONFIG_I2C_PXA_MODULE */

	/* LEDs */
#if defined(CONFIG_LEDS_PWM) || defined(CONFIG_LEDS_PWM_MODULE)
	GPIO20_PWM3_OUT, /* SODIMM pin 59, EvalBoard X11 A23 */
	GPIO19_PWM2_OUT, /* SODIMM pin 28, EvalBoard X8 A2 */
#endif /* CONFIG_LEDS_PWM || CONFIG_LEDS_PWM_MODULE */
#if defined(CONFIG_LEDS_GPIO) || defined(CONFIG_LEDS_GPIO_MODULE)
	/* borrowed from PS2Top-Clock and PS2Bottom-Clock
	   as they are not widely used nowdays */
	GPIO31_GPIO,		/* PXA300 */
	GPIO77_GPIO,		/* PXA300 */
	GPIO9_2_GPIO,		/* PXA310 */
	GPIO8_2_GPIO,		/* PXA310 */
#endif /* CONFIG_LEDS_GPIO || CONFIG_LEDS_GPIO_MODULE */

#if defined(CONFIG_MMC_PXA) || defined(CONFIG_MMC_PXA_MODULE)
	/* MMC1 */
	GPIO3_MMC1_DAT0 | MFP_LPM_EDGE_BOTH,
	GPIO4_MMC1_DAT1 | MFP_LPM_EDGE_BOTH,
	GPIO5_MMC1_DAT2 | MFP_LPM_EDGE_BOTH,
	GPIO6_MMC1_DAT3 | MFP_LPM_EDGE_BOTH,
	GPIO7_MMC1_CLK,
	GPIO13_GPIO | MFP_LPM_EDGE_BOTH,	/* Card detect interrupt */
	GPIO14_MMC1_CMD | MFP_LPM_EDGE_BOTH,	/* CMD0 for slot 0 */
#endif /* CONFIG_MMC_PXA || CONFIG_MMC_PXA_MODULE */

#if defined(CONFIG_USB_OHCI_HCD) || defined(CONFIG_USB_OHCI_HCD_MODULE)
	/* USB */
	GPIO0_2_USBH_PEN,	/* USB Host Power Enable */
	GPIO1_2_USBH_PWR,	/* USB Host Overcurrent Control */
	GPIO98_GPIO,		/* USB device cable detect, active high */
//USBHPEN2, muxed with USB_P2_8 depends on UP2OCR[SEOS]
//GPIO103, USB_P2_8
#endif /* CONFIG_USB_OHCI_HCD || CONFIG_USB_OHCI_HCD_MODULE */

#if defined(CONFIG_KEYBOARD_GPIO) || defined(CONFIG_KEYBOARD_GPIO_MODULE)
	/* Keys for Android
	   Note: the Toradex Colibri Evaluation Board will require jumper cables from
	   X11 (GPIO Usage) to X21 (LED/Switches) wired as follows
	   Key		GPIO	X11	X21
	   Home		11	B44	15	SWITCH8
	   Menu		97	B48	19	BUTTON2
	   Back		96	B47	17	BUTTON1 */
	GPIO11_GPIO | MFP_LPM_EDGE_RISE, /* Home */
	GPIO97_GPIO | MFP_LPM_EDGE_RISE, /* Menu */
	GPIO96_GPIO | MFP_LPM_EDGE_RISE, /* Back */
#endif /* CONFIG_KEYBOARD_GPIO || CONFIG_KEYBOARD_GPIO_MODULE */

};


#if defined(CONFIG_AX88796)
#define COLIBRI_ETH_IRQ_GPIO	mfp_to_gpio(GPIO26_GPIO)

/*
 * Asix AX88796 Ethernet
 */
static struct ax_plat_data colibri_asix_platdata = {
	.flags		= 0, /* defined later */
	.wordlength	= 2,
};

static struct resource colibri_asix_resource[] = {
	[0] = {
		.start = PXA3xx_CS2_PHYS,
		.end   = PXA3xx_CS2_PHYS + (0x20 * 2) - 1,
		.flags = IORESOURCE_MEM,
	},
	[1] = {
		.start = gpio_to_irq(COLIBRI_ETH_IRQ_GPIO),
		.end   = gpio_to_irq(COLIBRI_ETH_IRQ_GPIO),
		.flags = IORESOURCE_IRQ | IRQF_TRIGGER_FALLING,
	}
};

static struct platform_device asix_device = {
	.name		= "ax88796",
	.id		= 0,
	.num_resources 	= ARRAY_SIZE(colibri_asix_resource),
	.resource	= colibri_asix_resource,
	.dev		= {
		.platform_data = &colibri_asix_platdata
	}
};

static mfp_cfg_t colibri_pxa300_eth_pin_config[] __initdata = {
	GPIO1_nCS2,			/* AX88796 chip select */
	GPIO26_GPIO | MFP_PULL_HIGH	/* AX88796 IRQ */
};

static void __init colibri_pxa300_init_eth(void)
{
	colibri_pxa3xx_init_eth(&colibri_asix_platdata);
	pxa3xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa300_eth_pin_config));
	platform_device_register(&asix_device);
}
#else
static inline void __init colibri_pxa300_init_eth(void) {}
#endif /* CONFIG_AX88796 */

#if defined(CONFIG_MMC_PXA) || defined(CONFIG_MMC_PXA_MODULE)
static struct pxamci_platform_data colibri_mci_platform_data = {
	.detect_delay_ms  = 200,
	.ocr_mask	  = MMC_VDD_32_33|MMC_VDD_33_34,
	.gpio_card_ro     = -1,
	.gpio_card_detect = mfp_to_gpio(MFP_PIN_GPIO13),
	.gpio_power       = -1
};
#endif /* CONFIG_MMC_PXA || CONFIG_MMC_PXA_MODULE */

#if defined(CONFIG_FB_PXA) || defined(CONFIG_FB_PXA_MODULE)
static mfp_cfg_t colibri_pxa300_lcd_pin_config[] __initdata = {
	GPIO17_PWM0_OUT, /* PWM backlight */
	GPIO39_GPIO,		/* Backlight */
	GPIO54_LCD_LDD_0,
	GPIO55_LCD_LDD_1,
	GPIO56_LCD_LDD_2,
	GPIO57_LCD_LDD_3,
	GPIO58_LCD_LDD_4,
	GPIO59_LCD_LDD_5,
	GPIO60_LCD_LDD_6,
	GPIO61_LCD_LDD_7,
	GPIO62_LCD_LDD_8,
	GPIO63_LCD_LDD_9,
	GPIO64_LCD_LDD_10,
	GPIO65_LCD_LDD_11,
	GPIO66_LCD_LDD_12,
	GPIO67_LCD_LDD_13,
	GPIO68_LCD_LDD_14,
	GPIO69_LCD_LDD_15,
	GPIO70_LCD_LDD_16,
	GPIO71_LCD_LDD_17,
	GPIO62_LCD_CS_N,
	GPIO72_LCD_FCLK,
	GPIO73_LCD_LCLK,
	GPIO74_LCD_PCLK,
	GPIO75_LCD_BIAS,
	GPIO76_LCD_VSYNC,
};

static void __init colibri_pxa300_init_lcd(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa300_lcd_pin_config));
}

#else
static inline void colibri_pxa300_init_lcd(void) {}
#endif /* CONFIG_FB_PXA || CONFIG_FB_PXA_MODULE */

#if defined(CONFIG_SND_AC97_CODEC) || defined(CONFIG_SND_AC97_CODEC_MODULE)
static mfp_cfg_t colibri_pxa310_ac97_pin_config[] __initdata = {
	GPIO24_AC97_SYSCLK,
	GPIO23_AC97_nACRESET,
	GPIO25_AC97_SDATA_IN_0,
	GPIO27_AC97_SDATA_OUT,
	GPIO28_AC97_SYNC,
	GPIO29_AC97_BITCLK
};

static inline void __init colibri_pxa310_init_ac97(void)
{
	/* no AC97 codec on Colibri PXA300 */
	if (!cpu_is_pxa310())
		return;

	pxa3xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa310_ac97_pin_config));
	pxa_set_ac97_info(NULL);
}
#else
static inline void colibri_pxa310_init_ac97(void) {}
#endif

void __init colibri_pxa300_init(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa300_pin_config));
	pxa_set_ffuart_info(NULL);
	pxa_set_btuart_info(NULL);
	pxa_set_stuart_info(NULL);
	colibri_pxa_init_evalboard();
	colibri_pxa_init_gpio_keys(PXA_ALL_CB);

	colibri_pxa300_init_eth();
	colibri_pxa_init_ohci(mfp_to_gpio(MFP_PIN_GPIO98));
	colibri_pxa3xx_init_nand();
	colibri_pxa300_init_lcd();
	colibri_pxa_init_lcd(mfp_to_gpio(MFP_PIN_GPIO39));
	colibri_pxa_init_rtc();
	colibri_pxa310_init_ac97();
	colibri_pxa_init_mmc(&colibri_mci_platform_data);
}

MACHINE_START(COLIBRI300, "Toradex Colibri PXA300")
	.phys_io	= 0x40000000,
	.io_pg_offst	= (io_p2v(0x40000000) >> 18) & 0xfffc,
	.boot_params	= COLIBRI_SDRAM_BASE + 0x100,
	.init_machine	= colibri_pxa300_init,
	.map_io		= pxa_map_io,
	.init_irq	= pxa3xx_init_irq,
	.timer		= &pxa_timer,
MACHINE_END
