/*
 *  arch/arm/mach-pxa/colibri-pxa320.c
 *
 *  Support for Toradex PXA320/310 based Colibri module
 *
 *  Daniel Mack <daniel@caiaq.de>
 *  Matthias Meier <matthias.j.meier@gmx.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <linux/interrupt.h>

#include <asm/mach-types.h>
#include <asm/sizes.h>
#include <asm/mach/arch.h>
#include <asm/mach/irq.h>

#include <mach/pxa3xx-regs.h>
#include <mach/mfp-pxa320.h>
#include <mach/colibri.h>
#include <mach/pxafb.h>
#include <mach/audio.h>
#include <mach/mmc.h>

#include "generic.h"
#include "devices.h"
#include "colibri.h"

static mfp_cfg_t colibri_pxa320_pin_config[] __initdata = {
	/* I2C */
#if defined(CONFIG_I2C_PXA) || defined(CONFIG_I2C_PXA_MODULE)
	GPIO32_I2C_SCL | MFP_LPM_EDGE_BOTH,
	GPIO33_I2C_SDA | MFP_LPM_EDGE_BOTH,
#endif /* CONFIG_I2C_PXA || CONFIG_I2C_PXA_MODULE */

	/* LEDs */
#if defined(CONFIG_LEDS_PWM) || defined(CONFIG_LEDS_PWM_MODULE)
	GPIO14_PWM3_OUT, /* SODIMM pin 59, EvalBoard X11 A23 */
	GPIO13_PWM2_OUT, /* SODIMM pin 28, EvalBoard X8 A2 */
#endif /* CONFIG_LEDS_PWM || CONFIG_LEDS_PWM_MODULE */
#if defined(CONFIG_LEDS_GPIO) || defined(CONFIG_LEDS_GPIO_MODULE)
	/* borrowed from PS2Top-Clock and PS2Bottom-Clock
	   as they are not widely used nowdays */
	GPIO42_GPIO, /* SODIMM pin 63, EvalBoard X11 A25 */
	GPIO75_GPIO, /* SODIMM pin 69, EvalBoard X11 A28 */
#endif /* CONFIG_LEDS_GPIO || CONFIG_LEDS_GPIO_MODULE */

#if defined(CONFIG_MMC_PXA) || defined(CONFIG_MMC_PXA_MODULE)
	/* MMC1 */
	GPIO18_MMC1_DAT0 | MFP_LPM_EDGE_BOTH,
	GPIO19_MMC1_DAT1 | MFP_LPM_EDGE_BOTH,
	GPIO20_MMC1_DAT2 | MFP_LPM_EDGE_BOTH,
	GPIO21_MMC1_DAT3 | MFP_LPM_EDGE_BOTH,
	GPIO22_MMC1_CLK,
	GPIO23_MMC1_CMD | MFP_LPM_EDGE_BOTH,	/* CMD0 for slot 0 */
	GPIO28_GPIO | MFP_LPM_EDGE_BOTH,	/* Card detect interrupt */
#endif /* CONFIG_MMC_PXA || CONFIG_MMC_PXA_MODULE */

#if defined(CONFIG_USB_OHCI_HCD) || defined(CONFIG_USB_OHCI_HCD_MODULE)
	GPIO2_2_USBH_PEN, /* USB host power enable */
	GPIO3_2_USBH_PWR, /* USB host overcurrent control */
	GPIO96_GPIO,	  /* USB device cable detect, active high */
//USBHPEN2, muxed with USB_P2_8 depends on UP2OCR[SEOS]
//GPIO101, USB_P2_8
#endif /* CONFIG_USB_OHCI_HCD || CONFIG_USB_OHCI_HCD_MODULE */

#if defined(CONFIG_KEYBOARD_GPIO) || defined(CONFIG_KEYBOARD_GPIO_MODULE)
	/* Keys for Android
	   Note: the Toradex Colibri Evaluation Board will require jumper cables from
	   X11 (GPIO Usage) to X21 (LED/Switches) wired as follows
	   Key		GPIO	X11	X21
	   Home		26	B44	15	SWITCH8
	   Menu		95	B48	19	BUTTON2
	   Back		94	B47	17	BUTTON1 */
	GPIO26_GPIO | MFP_LPM_EDGE_RISE, /* Home */
	GPIO95_GPIO | MFP_LPM_EDGE_RISE, /* Menu */
	GPIO94_GPIO | MFP_LPM_EDGE_RISE, /* Back */
#endif /* CONFIG_KEYBOARD_GPIO || CONFIG_KEYBOARD_GPIO_MODULE */
};

#if defined(CONFIG_AX88796) || defined(CONFIG_AX88796C)
#define COLIBRI_ETH_IRQ_GPIO	mfp_to_gpio(GPIO36_GPIO)
#endif /* CONFIG_AX88796 || CONFIG_AX88796C */

#if defined(CONFIG_AX88796)
/*
 * Asix AX88796 Ethernet
 */
static struct ax_plat_data colibri_asix_platdata = {
	.flags		= 0, /* defined later */
	.wordlength	= 2,
};

static struct resource colibri_asix_resource[] = {
	[0] = {
		.start	= PXA3xx_CS2_PHYS,
		/* 2 means address lines shifted by one */
		.end	= PXA3xx_CS2_PHYS + (0x20 * 2) - 1,
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		/* FIFO_SEL connected to address line 11 */
		.start	= PXA3xx_CS2_PHYS + (1 << 11),
		.end	= PXA3xx_CS2_PHYS + (1 << 11),
		.flags	= IORESOURCE_MEM,
	},
	[2] = {
		.start	= gpio_to_irq(COLIBRI_ETH_IRQ_GPIO),
		.end	= gpio_to_irq(COLIBRI_ETH_IRQ_GPIO),
		.flags	= IORESOURCE_IRQ | IORESOURCE_IRQ_LOWEDGE,
	}
};

static struct platform_device asix_device = {
	.name		= "ax88796",
	.id		= 0,
	.num_resources 	= ARRAY_SIZE(colibri_asix_resource),
	.resource	= colibri_asix_resource,
	.dev		= {
		.platform_data = &colibri_asix_platdata
	}
};
#endif /* CONFIG_AX88796 */

#if defined(CONFIG_AX88796C)
static struct resource ax88796c_resource[] = {
	[0] = {
		.start	= PXA3xx_CS2_PHYS,		/* Start of AX88796C base address */
		.end	= PXA3xx_CS2_PHYS + 0xFFFFF,	/* End of AX88796C base address */
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start	= gpio_to_irq(COLIBRI_ETH_IRQ_GPIO),
		.end	= gpio_to_irq(COLIBRI_ETH_IRQ_GPIO),
		.flags	= IORESOURCE_IRQ | IORESOURCE_IRQ_LOWEDGE,
	}
};

struct platform_device net_device_ax88796c = {
	.name		= "ax88796c",
	.id		= -1,
	.num_resources	= ARRAY_SIZE(ax88796c_resource),
	.resource	= ax88796c_resource,
};
#endif /* CONFIG_AX88796C */

#if defined(CONFIG_AX88796) || defined(CONFIG_AX88796C)
static mfp_cfg_t colibri_pxa320_eth_pin_config[] __initdata = {
	GPIO3_nCS2,			/* AX88796 chip select */
	GPIO36_GPIO | MFP_PULL_HIGH	/* AX88796 IRQ */
};

static void __init colibri_pxa320_init_eth(void)
{
	colibri_pxa3xx_init_eth(&colibri_asix_platdata);
	pxa3xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa320_eth_pin_config));

	if (system_rev < 0x020a) {
		/* before V2.0a old AX88796B is used */
#if defined(CONFIG_AX88796)
		platform_device_register(&asix_device);
#endif /* CONFIG_AX88796 */
	} else {
		/* starting with V2.0a new AX88796C is used */
#if defined(CONFIG_AX88796C)
		platform_device_register(&net_device_ax88796c);
#endif /* CONFIG_AX88796C */
	}
}
#else
static inline void __init colibri_pxa320_init_eth(void) {}
#endif /* CONFIG_AX88796 || CONFIG_AX88796C */

#if defined(CONFIG_MMC_PXA) || defined(CONFIG_MMC_PXA_MODULE)
static struct pxamci_platform_data colibri_mci_platform_data = {
	.detect_delay_ms  = 200,
	.ocr_mask	  = MMC_VDD_32_33|MMC_VDD_33_34,
	.gpio_card_ro     = -1,
	.gpio_card_detect = mfp_to_gpio(MFP_PIN_GPIO28),
	.gpio_power       = -1
};
#endif /* CONFIG_MMC_PXA || CONFIG_MMC_PXA_MODULE */


#if defined(CONFIG_FB_PXA) || defined(CONFIG_FB_PXA_MODULE)
static mfp_cfg_t colibri_pxa320_lcd_pin_config[] __initdata = {
	GPIO11_PWM0_OUT,        /* PWM, backlight, pulled up */
	GPIO6_2_LCD_LDD_0,
	GPIO7_2_LCD_LDD_1,
	GPIO8_2_LCD_LDD_2,
	GPIO9_2_LCD_LDD_3,
	GPIO10_2_LCD_LDD_4,
	GPIO11_2_LCD_LDD_5,
	GPIO12_2_LCD_LDD_6,
	GPIO13_2_LCD_LDD_7,
	GPIO63_LCD_LDD_8,
	GPIO64_LCD_LDD_9,
	GPIO65_LCD_LDD_10,
	GPIO66_LCD_LDD_11,
	GPIO67_LCD_LDD_12,
	GPIO68_LCD_LDD_13,
	GPIO69_LCD_LDD_14,
	GPIO70_LCD_LDD_15,
	GPIO71_LCD_LDD_16,
	GPIO72_LCD_LDD_17,
	GPIO14_2_LCD_FCLK,
	GPIO15_2_LCD_LCLK,
	GPIO16_2_LCD_PCLK,
	GPIO17_2_LCD_BIAS,
#if defined(CONFIG_COLIBRI_LCD_EDT_ET057011DH6) /* Backlight */
	GPIO14_GPIO,
#else
	GPIO49_GPIO,
#endif
};

void __init colibri_pxa320_init_lcd(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa320_lcd_pin_config));
}

#else
static inline void colibri_pxa320_init_lcd(void) {}
#endif /* CONFIG_FB_PXA || CONFIG_FB_PXA_MODULE */

#if	defined(CONFIG_SND_AC97_CODEC) || \
	defined(CONFIG_SND_AC97_CODEC_MODULE)
static mfp_cfg_t colibri_pxa320_ac97_pin_config[] __initdata = {
	GPIO34_AC97_SYSCLK,
	GPIO35_AC97_SDATA_IN_0,
	GPIO37_AC97_SDATA_OUT,
	GPIO38_AC97_SYNC,
	GPIO39_AC97_BITCLK,
	GPIO40_AC97_nACRESET
};

static inline void __init colibri_pxa320_init_ac97(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa320_ac97_pin_config));

	if (system_rev < 0x020a) {
		/* before V2.0a old NXP (formerly Philips) UCB1400 is used */
		pxa_device_ac97.name = "pxa2xx-ac97-ucb1400";
	} else {
		/* starting with V2.0a new Wolfson WM9715L (fully WM9712 compatible) is used */
	}
	pxa_set_ac97_info(NULL);
}
#else
static inline void colibri_pxa320_init_ac97(void) {}
#endif

/*
 * The following configuration is verified to work with the Toradex Orchid
 * carrier board
 */
static mfp_cfg_t colibri_pxa320_uart_pin_config[] __initdata = {
	/* UART 1 configuration (may be set by bootloader) */
	GPIO99_UART1_CTS,
	GPIO104_UART1_RTS,
	GPIO97_UART1_RXD,
	GPIO98_UART1_TXD,
	GPIO101_UART1_DTR,
	GPIO103_UART1_DSR,
	GPIO100_UART1_DCD,
	GPIO102_UART1_RI,

	/* UART 2 configuration */
	GPIO109_UART2_CTS,
	GPIO112_UART2_RTS,
	GPIO110_UART2_RXD,
	GPIO111_UART2_TXD,

	/* UART 3 configuration */
	GPIO30_UART3_RXD,
	GPIO31_UART3_TXD,
};

static void __init colibri_pxa320_init_uart(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa320_uart_pin_config));
}

void __init colibri_pxa320_init(void)
{
	pxa3xx_mfp_config(ARRAY_AND_SIZE(colibri_pxa320_pin_config));
	pxa_set_ffuart_info(NULL);
	pxa_set_btuart_info(NULL);
	pxa_set_stuart_info(NULL);
	colibri_pxa_init_evalboard();
	colibri_pxa_init_gpio_keys(PXA_ALL_CB);


	colibri_pxa320_init_eth();
	colibri_pxa_init_ohci(mfp_to_gpio(MFP_PIN_GPIO96));
	colibri_pxa3xx_init_nand();
	colibri_pxa320_init_lcd();
#if defined(CONFIG_COLIBRI_LCD_EDT_ET057011DH6)
#define colibri_gpio_backlight mfp_to_gpio(MFP_PIN_GPIO14)
#else
#define colibri_gpio_backlight mfp_to_gpio(MFP_PIN_GPIO49)
#endif
	colibri_pxa_init_lcd(colibri_gpio_backlight);
	colibri_pxa_init_rtc();

	colibri_pxa320_init_ac97();
	colibri_pxa_init_mmc(&colibri_mci_platform_data);
	colibri_pxa320_init_uart();

}

MACHINE_START(COLIBRI320, "Toradex Colibri PXA320")
	.phys_io	= 0x40000000,
	.io_pg_offst	= (io_p2v(0x40000000) >> 18) & 0xfffc,
	.boot_params	= COLIBRI_SDRAM_BASE + 0x100,
	.init_machine	= colibri_pxa320_init,
	.map_io		= pxa_map_io,
	.init_irq	= pxa3xx_init_irq,
	.timer		= &pxa_timer,
MACHINE_END
