/*
 *  linux/arch/arm/mach-pxa/colibri-pxa270.c
 *
 *  Generic Toradex Colibri PXA270/PXA3xx functions
 *  Yury Georgievskiy <yury.georgievskiy@noser.com>
 *
 *  Copyright (C) 2011 Noser Engineering
 *
 *  2011-01-17: Yury Georgievskiy <yury.georgievskiy@noser.com>
 *              initial version
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <asm/irq.h>
#include <asm/mach-types.h>
#include <asm/mach/arch.h>
#include <asm/mach/flash.h>
#include <asm/mach/irq.h>
#include <asm/mach/map.h>
#include <asm/sizes.h>

#include <linux/bitops.h>
#include <linux/delay.h>
#include <linux/gpio.h>
#include <linux/gpio_keys.h>
#include <linux/i2c.h>
#include <linux/init.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/mtd/physmap.h>
#include <linux/platform_device.h>
#include <linux/sysdev.h>
#include <linux/usb/gpio_vbus.h>

#include <mach/colibri.h>
#include <mach/hardware.h>
#include <mach/mmc.h>
#include <mach/ohci.h>
#include <mach/pxa27x-udc.h>
#include <mach/udc.h>

#include <plat/i2c.h>

#include "colibri.h"

#if defined(CONFIG_RTC_DRV_DS1307) || defined(CONFIG_RTC_DRV_DS1307_MODULE)
static struct i2c_board_info __initdata colibri_i2c_rtc = {
	.type	= "m41t00",
	.addr	= 0x68,
};

void __init colibri_pxa_init_rtc(void)
{
	pxa_set_i2c_info(NULL);
	i2c_register_board_info(0, &colibri_i2c_rtc, 1);
}
#else
inline void __init colibri_pxa_init_rtc(void) {}
#endif /* #if CONFIG_RTC_DRV_DS1307 || CONFIG_RTC_DRV_DS1307_MODULE */

#if defined(CONFIG_MMC_PXA) || defined(CONFIG_MMC_PXA_MODULE)
void __init colibri_pxa_init_mmc(struct pxamci_platform_data *pd)
{
	pxa_set_mci_info(pd);
}
#else /* CONFIG_MMC_PXA || CONFIG_MMC_PXA_MODULE */
inline void __init colibri_pxa_init_mmc(struct pxamci_platform_data *pd) {}
#endif /* CONFIG_MMC_PXA || CONFIG_MMC_PXA_MODULE */

#if defined(CONFIG_USB_OHCI_HCD) || defined(CONFIG_USB_OHCI_HCD_MODULE)
static int colibri_ohci_init(struct device *dev)
{
	UP2OCR = UP2OCR_HXOE | UP2OCR_HXS; /* use port 2 as host */
	UP2OCR |= UP2OCR_DPPDE | UP2OCR_DMPDE; /* enable D +/- pull-down */

	return 0;
}

static struct pxaohci_platform_data colibri_ohci_platform_data = {
	.port_mode	= PMM_GLOBAL_MODE,	/* All ports powered at the
						   same time. */
	.power_budget	= 500,
	.init		= colibri_ohci_init,
	.flags		= ENABLE_PORT1 | ENABLE_PORT2 |
        /* Set the Power Control Polarity Low and Power Sense
           Polarity Low to active low. */
				POWER_CONTROL_LOW | POWER_SENSE_LOW,
};

#if defined(CONFIG_USB_GADGET_PXA27X) || \
    defined(CONFIG_USB_GADGET_PXA27X_MODULE)
static struct pxa2xx_udc_mach_info colibri_udc_info __initdata = {
	/* no D+ pullup; Colibri can't connect/disconnect in software */
	.gpio_pullup	= -1,
	/* gpio_vbus only used in pxa25x_udc driver */
};

static struct gpio_vbus_mach_info colibri_gpio_vbus_info = {
	.gpio_pullup		= -1,
	/* gpio_vbus set below */
};

static struct platform_device colibri_gpio_vbus = {
	.name	= "gpio-vbus",
	.id	= -1,
	.dev	= {
		.platform_data	= &colibri_gpio_vbus_info,
	},
};
#endif /* CONFIG_USB_GADGET_PXA27X) || CONFIG_USB_GADGET_PXA27X_MODULE */

//vbus sensing is really USB device and not host (e.g. OHCI) specific
void __init colibri_pxa_init_ohci(int gpio_vbus) /* VBUS sensing GPIO */
{
#if defined(CONFIG_USB_GADGET_PXA27X) || \
    defined(CONFIG_USB_GADGET_PXA27X_MODULE)
	pxa_set_udc_info(&colibri_udc_info);
#endif /* CONFIG_USB_GADGET_PXA27X || CONFIG_USB_GADGET_PXA27X_MODULE */
	pxa_set_ohci_info(&colibri_ohci_platform_data);
#ifdef MECS_TELLURIUM
	/* MECS Tellurium has a dedicated USB client/device port */
	colibri_gpio_vbus_info.gpio_vbus = -1;
#else
	colibri_gpio_vbus_info.gpio_vbus = gpio_vbus;
#endif
	platform_device_register(&colibri_gpio_vbus);
}
#else
inline void __init colibri_pxa_init_ohci(int gpio_vbus) {}
#endif /* CONFIG_USB_OHCI_HCD || CONFIG_USB_OHCI_HCD_MODULE */

#if defined(CONFIG_KEYBOARD_GPIO) || defined(CONFIG_KEYBOARD_GPIO_MODULE)

enum {
	_K_HOME = 0,
	_K_MENU,
	_K_BACK
};

static struct gpio_keys_button colibri_gpio_keys[] = {
	[_K_HOME] {
		.active_low	= 0,
		.code		= KEY_HOME,
		.desc		= "Home Button",
		.type		= EV_KEY,
		.wakeup		= 1,
		.debounce_interval = 50,
	},
	[_K_MENU] {
		.active_low	= 0,
		.code		= KEY_KBDILLUMDOWN,
		.desc		= "Menu Button",
		.type		= EV_KEY,
		.wakeup		= 1,
		.debounce_interval = 50,
	},
	[_K_BACK] {
		.active_low	= 0,
		.code		= KEY_BACK,
		.desc		= "Back Button",
		.type		= EV_KEY,
		.wakeup		= 1,
		.debounce_interval = 50,
	},
};

static struct gpio_keys_platform_data colibri_gpio_keys_platform_data = {
	.buttons	= colibri_gpio_keys,
	.nbuttons	= ARRAY_SIZE(colibri_gpio_keys),
};

static struct platform_device colibri_gpio_keys_device = {
	.name	= "gpio-keys",
	.id	= -1,
	.dev	= {
		.platform_data = &colibri_gpio_keys_platform_data,
	},
};

void __init colibri_pxa_init_gpio_keys(pxa_cb_t cb)
{
	if (cpu_is_pxa320()) {
		if (cb == PXA_LIMESTONE_CB) {
			colibri_gpio_keys[_K_HOME].gpio =
				mfp_to_gpio(MFP_PIN_GPIO24);
			colibri_gpio_keys[_K_MENU].gpio =
				mfp_to_gpio(MFP_PIN_GPIO25);
			colibri_gpio_keys[_K_BACK].gpio =
				mfp_to_gpio(MFP_PIN_GPIO26);
		} else {
			colibri_gpio_keys[_K_HOME].gpio =
				mfp_to_gpio(MFP_PIN_GPIO26);
			colibri_gpio_keys[_K_MENU].gpio =
				mfp_to_gpio(MFP_PIN_GPIO95);
			colibri_gpio_keys[_K_BACK].gpio =
				mfp_to_gpio(MFP_PIN_GPIO94);
		}
	} else if (cpu_is_pxa27x()) {
		colibri_gpio_keys[_K_HOME].gpio = mfp_to_gpio(MFP_PIN_GPIO36);
		colibri_gpio_keys[_K_MENU].gpio = mfp_to_gpio(MFP_PIN_GPIO35);
		colibri_gpio_keys[_K_BACK].gpio = mfp_to_gpio(MFP_PIN_GPIO37);
	} else if (cpu_is_pxa300()) {
		colibri_gpio_keys[_K_HOME].gpio = mfp_to_gpio(MFP_PIN_GPIO11);
		colibri_gpio_keys[_K_MENU].gpio = mfp_to_gpio(MFP_PIN_GPIO97);
		colibri_gpio_keys[_K_BACK].gpio = mfp_to_gpio(MFP_PIN_GPIO96);
	} else if (cpu_is_pxa310()) {
		colibri_gpio_keys[_K_HOME].gpio = mfp_to_gpio(MFP_PIN_GPIO11);
		colibri_gpio_keys[_K_MENU].gpio = mfp_to_gpio(MFP_PIN_GPIO97);
		colibri_gpio_keys[_K_BACK].gpio = mfp_to_gpio(MFP_PIN_GPIO96);
	} else {
		panic("Unsupported machine: %d\n", machine_arch_type);
	}

	platform_device_register(&colibri_gpio_keys_device);
}
#else /* CONFIG_KEYBOARD_GPIO || CONFIG_KEYBOARD_GPIO_MODULE */
inline void colibri_pxa_init_gpio_keys(void) {}
#endif /* CONFIG_KEYBOARD_GPIO || CONFIG_KEYBOARD_GPIO_MODULE */
